#include <WiFi.h>
#include <HTTPClient.h>
#include <EEPROM.h>
#include <WebServer.h>
#include <ArduinoJson.h>
#include <time.h> 
#include <esp_task_wdt.h>  // Biblioteca do Watchdog

// Configuração da rede Wi-Fi
const char* ssid = "Labmaker";
const char* password = "Anglo2020pp";

// Definição dos pinos e configuração inicial
#define RELAY_PIN 23
#define led_pin 25

// estrutura que será sala no array das programações de hora
struct Programacao {
  int horaLiga;
  int minutoLiga;
  int horaDesliga;
  int minutoDesliga;
  int diasSemana;
};

Programacao programacoes[5]; // criando um vetor de 5 posições com a estrutura 
WebServer server(80); // porta ado webserver que ficará o HTML para acionamento do painel
tm horarioAtual = {};
time_t ultimaSincronizacao;  

void atualizarRelogio(tm &timeInfo) {
  time_t t = mktime(&timeInfo);
  timeval now = { .tv_sec = t };
  settimeofday(&now, nullptr);
  ultimaSincronizacao = time(nullptr);
}

// Função para salvar as programações na EEPROM
void salvarProgramacoes() { 
  for (int i = 0; i < 5; i++) {
    EEPROM.put(i * sizeof(Programacao), programacoes[i]);
  }
  EEPROM.commit();
}

// Função para carregar as programações da EEPROM
void carregarProgramacoes() {
  for (int i = 0; i < 5; i++) {
    EEPROM.get(i * sizeof(Programacao), programacoes[i]);
  }
}

tm obterHorarioAtual() {   // carrega o horario atualizado (ao ligar no setup e ao definir um tempo de sincronia dentro do loop)
  WiFiClient client;
  HTTPClient http;
  tm timeInfo = {};

  http.begin(client, "http://worldtimeapi.org/api/timezone/America/Sao_Paulo");
  int httpCode = http.GET();

  if (httpCode == HTTP_CODE_OK) {
    String payload = http.getString();
    DynamicJsonDocument doc(1024);
    deserializeJson(doc, payload);
    String dateTime = doc["datetime"]; // Formato ISO 8601: "2024-08-01T10:00:40.903823-03:00"

    strptime(dateTime.c_str(), "%Y-%m-%dT%H:%M:%S", &timeInfo);
    atualizarRelogio(timeInfo);
  }

  http.end();
  return timeInfo;
}

// Função para verificar se um dia da semana está marcado
bool isDayChecked(int diasSemana, int day) { // para salvar os dias da semana utilizamos bits 0 e 1 para cada dia da semana iniciando no domingo. 
  return (diasSemana & (1 << day)) != 0;
}

// Função para verificar se o relé deve estar ligado
bool deveLigarRele(tm horarioAtual) {
  int diaSemanaAtual = horarioAtual.tm_wday;
  int horaAtual = horarioAtual.tm_hour;
  int minutoAtual = horarioAtual.tm_min;

  for (int i = 0; i < 5; i++) {
    if (isDayChecked(programacoes[i].diasSemana, diaSemanaAtual)) { // verifica se estamos no dia da semana correto
      if (horaAtual > programacoes[i].horaLiga || // verifica se a hora e os minutos é maior ou igual a hora programada
         (horaAtual == programacoes[i].horaLiga && minutoAtual >= programacoes[i].minutoLiga)) {
        if (horaAtual < programacoes[i].horaDesliga || //e se for maior, verifica se é menor a hora e os minutos para desligar
           (horaAtual == programacoes[i].horaDesliga && minutoAtual < programacoes[i].minutoDesliga)) {
          return true; // se sim habilita o acionamento do relé
        }
      }
    }
  }
  return false; // se nao entrar nas codições acima, retorna falso
}

String formatarDoisDigitos(int valor) { // para exibir no navegador as horas e os minutos menores que 10, precisamos acrescentar o 0 antes. aqui ele verifica e retorna o valor com o 0 se for menor que 10
  if (valor < 10) {
    return "0" + String(valor);
  } else {
    return String(valor);
  }
}

// Manipulador para a página principal
void handleRoot() { // pagina html 
  String page = "<!DOCTYPE html><html><head><meta charset='UTF-8'><title>Controle de LED</title>"; // charset='UTF-8' faz que os caracteres especais() ç ^ ~) apareçam de forma correta no HTML
  page += "<style>body {font-size: 24px; text-align: center;} table {margin: auto; border-collapse: collapse;} th, td {padding: 5px;} input[type='number'] {width: 50px;} tr {line-height: 1;}</style>";

              // aqui temos o formulario json que atualiza de forma dinamica partes da pagina. 
  page += "<script>setInterval(() => { fetch('/data').then(res => res.json()).then(data => { document.getElementById('dateTime').innerText = data.dateTime; document.getElementById('releState').innerText = data.releState ? 'Relé ligado' : 'Relé desligado'; document.getElementById('releState').style.color = data.releState ? 'red' : 'green'; document.getElementById('lastSync').innerText = data.lastSync; }); }, 1000);</script>";
  page += "</head><body><h1>Programação Painel Cantina</h1>"; // 
  page += "<form action='/salvar' method='POST'>";    //
  page += "<table border='1'><tr><th>Hora Liga</th><th>Hora Desliga</th><th>Dias da Semana</th></tr>";
    // definindo cabeçalho antes de carregar os valores que estao configurados na eeprom. 
  String diasSemana[7] = {"D", "S", "T", "Q", "Q", "S", "S"};

  for (int i = 0; i < 5; i++) { // como definimos um vetor de 5 posições, adiciona ao html as linhas correspondentes de cada programação salva. se precisar aumentar a quantidade de programações...
                                // basta mudar o limite da repetição na manipulação do vetor.
     page += "<tr>";
    page += "<td><input type='time' name='horaLiga" + String(i) + "' value='" + formatarDoisDigitos(programacoes[i].horaLiga) + ":" + formatarDoisDigitos(programacoes[i].minutoLiga) + "' step='60'></td>";
    page += "<td><input type='time' name='horaDesliga" + String(i) + "' value='" + formatarDoisDigitos(programacoes[i].horaDesliga) + ":" + formatarDoisDigitos(programacoes[i].minutoDesliga) + "' step='60'></td>";
     page += "<td>";
    page += "<table><tr>";
    for (int day = 0; day < 7; day++) { 
      page += "<th>" + diasSemana[day] + "</th>"; // insere os dias da semana na tabela
    }
    page += "</tr><tr>";
    for (int day = 0; day < 7; day++) { // insere os valores que foram habilitados e carregados da eeprom
      page += "<td><input type='checkbox' name='diasSemana" + String(i) + "_" + String(day) + "' value='1' " + (isDayChecked(programacoes[i].diasSemana, day) ? "checked" : "") + "></td>";
    }
    page += "</tr></table>";
    page += "</td>";
    page += "</tr>";
  }

  page += "</table>";
  page += "<p><button type='submit'>Salvar</button></p>"; // ao realizar as alterações, clicando em salvar, o programa chama a função que vai salvar os valores no vetor e posteriormente na eeprom
  page += "</form>";

  char dateTimeBuffer[20];
  strftime(dateTimeBuffer, sizeof(dateTimeBuffer), "%Y-%m-%d %H:%M:%S", &horarioAtual);
  String dateTimeAtual = String(dateTimeBuffer);
// aqui temos uma atualização dinamica apenas destes 3 campos sem afetar o restante da pagina
  page += "<p>Data e Hora Atual: <span id='dateTime'>" + dateTimeAtual + "</span></p>"; // exibe a data atualziando os segundos no html
  page += "<p>Última Sincronização: <span id='lastSync'></span></p>"; // exibe quando atualizamos o horario pela ultima vez.
  page += "<p id='releState'></p>"; // exibe o estado do relé (ligado ou desligado)

  page += "</body></html>";

  server.send(200, "text/html", page);
}

// Manipulador para a rota de salvamento
void handleSalvar() {
  for (int i = 0; i < 5; i++) {
    if (server.hasArg("horaLiga" + String(i))) {
      String horaLigaStr = server.arg("horaLiga" + String(i));
      programacoes[i].horaLiga = horaLigaStr.substring(0, 2).toInt();
      programacoes[i].minutoLiga = horaLigaStr.substring(3, 5).toInt();
    }
    if (server.hasArg("horaDesliga" + String(i))) {
      String horaDesligaStr = server.arg("horaDesliga" + String(i));
      programacoes[i].horaDesliga = horaDesligaStr.substring(0, 2).toInt();
      programacoes[i].minutoDesliga = horaDesligaStr.substring(3, 5).toInt();
    }
    
    int diasSemana = 0;
    for (int day = 0; day < 7; day++) {
      if (server.hasArg("diasSemana" + String(i) + "_" + String(day))) {
        diasSemana |= (1 << day);
      }
    }
    programacoes[i].diasSemana = diasSemana;
  }

  salvarProgramacoes(); // chama a função que salva na eeprom

                    // exibe na proxima pagina que as as programações foram salvas, ai clicando em voltar, exibe a programação novamente. 
  String page = "<!DOCTYPE html><html><head><meta charset='UTF-8'><title>Programações Salvas</title></head><body><h1>Programações salvas!</h1><a href='/'>Voltar</a></body></html>";

  server.send(200, "text/html", page);
}

// Manipulador para a rota de dados
void handleData() {
  tm horarioAtual;
  getLocalTime(&horarioAtual);
  char dateTimeBuffer[20];
  strftime(dateTimeBuffer, sizeof(dateTimeBuffer), "%Y-%m-%d %H:%M:%S", &horarioAtual);
  String dateTimeAtual = String(dateTimeBuffer);

  char lastSyncBuffer[20];
  tm *tmSync = localtime(&ultimaSincronizacao);
  strftime(lastSyncBuffer, sizeof(lastSyncBuffer), "%Y-%m-%d %H:%M:%S", tmSync);
  String lastSyncTime = String(lastSyncBuffer);

  bool ligarRele = deveLigarRele(horarioAtual);

  DynamicJsonDocument doc(1024);
  doc["dateTime"] = dateTimeAtual;
  doc["releState"] = ligarRele;
  doc["lastSync"] = lastSyncTime;

  String json;
  serializeJson(doc, json);
  server.send(200, "application/json", json);
}

void setup() {
  Serial.begin(115200);
  EEPROM.begin(512);
  pinMode(RELAY_PIN, OUTPUT);
  pinMode(led_pin, OUTPUT);

  digitalWrite(RELAY_PIN, LOW);

  carregarProgramacoes(); // carrega valores salvos da eeprom

  WiFi.begin(ssid, password); // inicia conexao no wifi
  while (WiFi.status() != WL_CONNECTED) {
    delay(1000);
    Serial.println("Conectando ao Wi-Fi...");
  }
  Serial.print("Conectado ao Wi-Fi - IP: ");
  Serial.println(WiFi.localIP());

  horarioAtual = obterHorarioAtual(); // obtem o horario atual 

  server.on("/", handleRoot);
  server.on("/salvar", HTTP_POST, handleSalvar);
  server.on("/data", handleData);
  server.begin();
  Serial.println("Servidor HTTP iniciado");


   // Configurar e iniciar o watchdog com um tempo limite de 10 segundos
  esp_task_wdt_config_t wdt_config = {
    .timeout_ms = 10000,  // Tempo limite em milissegundos
    .idle_core_mask = 0,  // Ativar para todos os núcleos
    .trigger_panic = true // Acionar um panic se o WDT expirar
  };
  esp_task_wdt_init(&wdt_config);  // Inicializa o WDT com a configuração fornecida
  esp_task_wdt_add(NULL);  // Adiciona a tarefa atual (loop) ao WDT
}

void loop() {
  esp_task_wdt_reset();  // Reseta o watchdog a cada loop para evitar reinicializações
  server.handleClient();

  static unsigned long lastSyncCheck = 0;
  unsigned long now = millis();
  
  if (now - lastSyncCheck >= (15*60*60*1000)) { // Verifica a cada 15 horas/ coloquei este calculo para nao exibir o valor todo em milissegundos. assim sei exatamente o tempo de sincronia em horas)
    lastSyncCheck = now;
    horarioAtual = obterHorarioAtual();
  }

  getLocalTime(&horarioAtual);
  bool ligarRele = deveLigarRele(horarioAtual);

  if (ligarRele && digitalRead(RELAY_PIN)== LOW) { // aciona o relé apenas uma vez, que é quando ligarele == true e o relé esteja desligado 
    digitalWrite(RELAY_PIN, HIGH);
    digitalWrite(led_pin, HIGH);
    Serial.println("Relé ligado");
  } else if (!ligarRele && digitalRead(RELAY_PIN) == HIGH){// desliga o relé apenas uma vez, que é quando ligarele == false e o relé esteja ligado
    digitalWrite(RELAY_PIN, LOW);
    digitalWrite(led_pin, LOW);
    Serial.println("Relé desligado");
  }
}
